@@ -32,6 +32,7 @@ class Document:
     FACT_NAME_PHONE_STRICT = "PHONE_strict"
     FACT_NAMES_NER = ("PER", "ORG", "GPE", "LOC")
     FACT_NAME_NAMEMAIL = "NAMEMAIL"
+    FACT_NAME_BOUNDED = "BOUNDED"
 
     KNOWN_ENTITIES = (FACT_NAME_EMAIL, FACT_NAME_ADDRESS, FACT_NAME_PHONE_STRICT, FACT_NAME_NAMEMAIL) + FACT_NAMES_NER
 
@@ -159,8 +160,14 @@ class Document:
         Retrieves list-based entities.
         """
         text = self.get_words()
+        lemmas = self.get_lemma()
+
         hits = self.entity_mapper.map_entities(text)
-        for entity_type, entity_values in hits.items():
+        lemma_hits = self.entity_mapper.map_entities(lemmas, entity_types=["CURRENCY"])
+        # combine hits
+        all_hits = z = {**hits, **lemma_hits}
+        # make facts
+        for entity_type, entity_values in all_hits.items():
             for entity_value in entity_values:
                 new_fact = Fact(
                     fact_type=entity_type,
@@ -169,7 +176,36 @@ class Document:
                     spans=[[entity_value["span"][0], entity_value["span"][1]]]
                 )
                 self.__texta_facts.append(new_fact)
+        # declare the entities processed
+        self.entities_processed = True
+        
 
+    def currency_sum(self):
+        """
+        Extracts currency + sum and sum + currency patterns from text using regexp.
+        Saves extractions as facts.
+        """
+        text = self.get_words()
+        currency_facts = [fact for fact in self.__texta_facts if fact.fact_type == "CURRENCY"]
+        for fact in currency_facts:
+            regexes = (
+                f"{fact.fact_value} [0-9,\.]+",
+                f"[0-9,\.]+ {fact.fact_value}[a-z]*"
+            )
+            for currency_regex in regexes:
+                pattern = re.compile(currency_regex)
+                for match in pattern.finditer(text):
+                    fact_value = match.string[match.start():match.end()]
+                    # recheck that string contains a number
+                    if any(map(str.isdigit, fact_value)):
+                        new_fact = Fact(
+                            fact_type="CURRENCY_SUM",
+                            fact_value=fact_value,
+                            doc_path=self.doc_path,
+                            spans=[match.start(), match.end()]
+                        )
+                        self.__texta_facts.append(new_fact)
+    
 
     def emails(self):
         text = self.get_words()
@@ -243,7 +279,9 @@ class Document:
                     j = i
                     for entity_word in splitted[1:]:
                         j += 1
-                        if entity_word != self.stanza_words[j].text:
+                        if j < len(self.stanza_words) and entity_word != self.stanza_words[j].text:
+                            isthatphrase = False
+                        if j >= len(self.stanza_words):
                             isthatphrase = False
                     if isthatphrase:
                         lemmas += word.lemma
@@ -313,7 +351,7 @@ class Document:
             fact.fact_value = fact.fact_value.strip("><\)\(:;-\.,\!\?")
         for fact in facts:
             facts_values += [fact.fact_value]
-            if fact.fact_type == "NAMEMAIL":
+            if fact.fact_type == Document.FACT_NAME_NAMEMAIL:
                 splitted_value = fact.fact_value.split(" ")
                 name_fact_value = " ".join(splitted_value[:2])
                 name_fact = Fact(
@@ -336,7 +374,7 @@ class Document:
                 ends += [email_fact.spans[0][1], name_fact.spans[0][1]]
 
         for fact in facts:
-            if fact.fact_type != "NAMEMAIL" and fact.spans[0][0] not in starts and fact.spans[0][1] not in ends:
+            if fact.fact_type != Document.FACT_NAME_NAMEMAIL and fact.spans[0][0] not in starts and fact.spans[0][1] not in ends:
                 fact.fact_value = fact.fact_value.strip("><\)\(:;-\.,\!\?")
                 new_facts += [fact]
         new_facts_values = list()
@@ -361,31 +399,31 @@ class Document:
             issubitem = False
             for j, item2 in enumerate(bound.items()):
                 if i != j and i not in no_is and j not in no_is:
-                    if re.search("(^|\s)" + re.escape(item1[0].strip("><\)\(:;-\.,\!\?")) + "(\s|$|\@[a-z]{4,5}\.[a-z]{2})", item2[0].strip("><\)\(:;-\.,\!\?")) and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(item2[0].strip("><\)\(:;-\.,\!\?")):
+                    if re.search("(^|\s)" + re.escape(item1[0].strip("><)(:;-.,!?")) + "(\s|$|\@[a-z]{4,5}\.[a-z]{2})", item2[0].strip("><)(:;-.,!?")) and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(item2[0].strip("><\)\(:;-\.,\!\?")):
                         if item2[1] in new_bound:
-                            new_bound[item2[1]] += [item2[0].strip("><\)\(:;-\.,\!\?")]
+                            new_bound[item2[1]] += [item2[0].strip("><)(:;-.,!?")]
                         else:
-                            new_bound[item2[1]] = [item2[0].strip("><\)\(:;-\.,\!\?")]
+                            new_bound[item2[1]] = [item2[0].strip("><)(:;-.,!?")]
                         no_is += [i]
                         issubitem = True
                     else:  # each word of this entity exists also in the other one.
                         notsameparts = False
                         for part_of_entity in item1[0].split(" "):
-                            if not re.search("(^|\s)" + re.escape(part_of_entity.strip("><\)\(:;-\.,\!\?")) + "(\s|$)", item2[0].strip("><\)\(:;-\.,\!\?")):
+                            if not re.search("(^|\s)" + re.escape(part_of_entity.strip("><)(:;-.,!?")) + "(\s|$)", item2[0].strip("><)(:;-.,!?")):
                                 notsameparts = True
                                 break
-                        if not notsameparts and len(item1[0].strip("><\)\(:;-\.,\!\?")) <= len(item2[0].strip("><\)\(:;-\.,\!\?")):
+                        if not notsameparts and len(item1[0].strip("><)(:;-.,!?")) <= len(item2[0].strip("><)(:;-.,!?")):
                             if item2[1] in new_bound:
-                                new_bound[item2[1]] += [re.sub("(\'|\")", "\*", item2[0].strip("><\)\(:;-\.,\!\?"))]
+                                new_bound[item2[1]] += [re.sub("(\'|\")", "*", item2[0].strip("><)(:;-.,!?"))]
                             else:
-                                new_bound[item2[1]] = [re.sub("(\'|\")", "\*", item2[0].strip("><\)\(:;-\.,\!\?"))]
+                                new_bound[item2[1]] = [re.sub("(\'|\")", "*", item2[0].strip("><)(:;-.,!?"))]
                             no_is += [i]
                             issubitem = True
             if not issubitem:
                 if item1[1] in new_bound:
-                    new_bound[item1[1]] += [re.sub("(\'|\")", "\*", item1[0].strip("><\)\(:;-\.,\!\?"))]
+                    new_bound[item1[1]] += [re.sub("(\'|\")", "*", item1[0].strip("><\)\(:;-\.,\!\?"))]
                 else:
-                    new_bound[item1[1]] = [re.sub("(\'|\")", "\*", item1[0].strip("><\)\(:;-\.,\!\?"))]
+                    new_bound[item1[1]] = [re.sub("(\'|\")", "*", item1[0].strip("><\)\(:;-\.,\!\?"))]
         for entity_type in new_bound:
             new_bound[entity_type] = list(set(new_bound[entity_type]))
         return new_bound
@@ -424,7 +462,7 @@ class Document:
         Bound facts found in that email, that has distance of  <=self.CLOSE_FACT_DISTANCE and that has self.space_between_ok True, together.
         """
         facts = sorted(facts, key=lambda item: item.spans[0][0])
-        new_bound = {'doc_path': self.doc_path, 'fact': 'BOUNDED', 'spans': [], 'str_val': {}, "str_values": []}
+        new_bound = {'doc_path': self.doc_path, 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {}, "str_values": []}
         new_bounded = []
         for fact in facts:
             if len(new_bound["spans"]) == 0:
@@ -469,7 +507,7 @@ class Document:
                             to_delete_is += [j]
                         elif span1[1] < span2[1]:
                             to_delete_is += [i]
-            new_bound_erased_coinciding = {'doc_path': self.doc_path, 'fact': 'BOUNDED', 'spans': [], 'str_val': {}}
+            new_bound_erased_coinciding = {'doc_path': self.doc_path, 'fact': Document.FACT_NAME_BOUNDED, 'spans': [], 'str_val': {}}
             for i, span1 in enumerate(new_bound["spans"]):
                 if i not in to_delete_is:
                     new_bound_erased_coinciding["spans"] += [span1]
@@ -525,12 +563,13 @@ class Document:
         return new_bounds
 
 
-    def add_phone_high_recall(self, facts):
+    def remove_old_bounded_and_add_phone_high_recall(self, facts):
         """
+        remove previous BOUNDED facts
         remove previous phone facts
         add phoneparser made for emails (ContactPhoneParserHighRecall)
         """
-        new_facts = [fact for fact in facts if not re.match("PHONE", fact.fact_type)]
+        new_facts = [fact for fact in facts if not re.match("PHONE|"+Document.FACT_NAME_BOUNDED, fact.fact_type)]
         text = self.get_words()
         phone_numbers = ContactPhoneParserHighRecall(text, months=self.concat_resources["months"]).parse()
         new_facts.extend((number.to_fact(Document.FACT_NAME_PHONE_HIGH_RECALL, self.doc_path) for number in phone_numbers))
@@ -539,7 +578,7 @@ class Document:
 
     def bounded(self):
         facts = self.remove_duplicate_facts_by_span(self.__texta_facts)
-        facts = self.add_phone_high_recall(facts)  # bounded uses only ContactPhoneParserHighRecall results.
+        facts = self.remove_old_bounded_and_add_phone_high_recall(facts)  # bounded uses only ContactPhoneParserHighRecall results and doesn't look at old BOUNDED facts.
         new_bounded = self.bound_close_ones(facts)
         new_bounded = self.remove_overlaping_in_bounded(new_bounded)
         new_bounded = self.concatenate_subset_bounds(new_bounded)
